COMMENT *

	File:	JBOOT.ASM	Created	6-JUN-1995	Richard B. Johnson
	Purpose:
	This program loads a compressed Linux kernel image into memory
	and starts its execution.

	It is a smaller and more robust version of a similar boot loader
	written by F. Coutant in 1992. He did most of the hard work of
	finding out where the stuff goes in memory.

	To overcome the limitation that this code and stack must all
	exist below segment 1000 hex because the kernel must be loaded
	there, I check to see if there is a color screen card. If so,
	I relocate all the code to the second page of screen memory
	and execute it from there!  Yes, RAM is RAM. If there is not a
	screen card, the code is executed from wherever DOS loaded it.

	It is assumed that the Linux kernel file will be placed on a
	hard disk, not a bootable floppy. Therefore, it is expected that
	if the file can be found, it will be read entirely.

	It is further assumed that JBOOT.COM will replace the shell
	in CONFIG.SYS to force Linux to be loaded:

	[MENU]
	MENUITEM=NORMAL, DOS Operating System
	MENUITEM=LINUX, Linux Operating System
	MENUCOLOR=7,1
	MENUDEFAULT=LINUX, 4

	[LINUX]
	STACKS=0,0
	FCBS=1,0
	FILES=8
	BUFFERS=1,0
	SHELL=C:\LBOOT\JBOOT.COM C:\LBOOT\VMLINUX

	[NORMAL]
	DOS=HIGH
	STACKS=0,0
	BREAK=ON
	DEVICE=C:\DOS\HIMEM.SYS /TESTMEM:OFF
	DEVICE=C:\DOS\EMM386.EXE RAM X=C800-CFFF X=D000-D7FF
	BUFFERS=16,0
	FILES=32
	DOS=UMB
	LASTDRIVE=M
	FCBS=6,0
	DEVICEHIGH /L:1,12048 =C:\DOS\SETVER.EXE
	SHELL=C:\DOS\COMMAND.COM C:\DOS\ /E:768 /p

	Notice:
	JBOOT is not intended to be a Copyright Work. It is donated
	to the Linux gurus worldwide. It may not even be useful and
	is not guaranteed to perform as designed. JBOOT may be used
	for any purpose and copied freely. It is not necessary to
	provide any credits to the author for any derived works.

	However, if you make any improvements, please email me patches
	at rjohnson@analogic.com.

*

.286

;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;	This defines where the kernel is finally loaded.
;
KERNEL	SEGMENT	AT	1000H
KERNEL	ENDS
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;	This is a template of the setup area.
;
SETUP	SEGMENT	AT	9000H
	ORG	1FEH			; Offset to signature
SIGNAT	DW	?			; Signature word
KSTART	LABEL	BYTE			; 2k of startup code
	ORG	4000H			; Offset to stack
KER_SP	DW	?			; Kernel start stack pointer
SETUP	ENDS
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;	This is a template of the data area used by the MS-DOS memory
;	manager if it is installed.
;
VECTABL	STRUC
INT_NO	DB	?			; Interrupt number
VCT_OFF	DW	?			; Vector offset
VCT_SEG	DW	?			; Vector segment
VECTABL	ENDS
;
MEMMAN	SEGMENT	AT	70H
	ORG	100H			; Start of saved interrupt table
SAVED	VECTABL	<>
MEMMAN	ENDS
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;	This is a template of the screen regen buffer. If this exists,
;	we are going to relocate the startup code to the screen buffer
;	and execute it from there!
;
SCREEN	SEGMENT	AT	0B800H
PAGE_ZERO	LABEL	BYTE
	ORG	1000H			; Bypass page 0
PAGE_ONE	LABEL	BYTE		; Where to relocate code
SCREEN	ENDS

;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
PAGE_LEN	EQU	4000H		; This must be a power of 2
PAGE_SEG	EQU	(PAGE_LEN SHR 4); Segment offset to new page
SETUPLEN	EQU	800H		; Length of the setup code
DOS_OPEN	EQU	3D00H		; MS-DOS Open for READ function
DOS_CLOSE	EQU	3E00H		; MS-DOS Close file function
DOS_READ	EQU	3F00H		; MS-DOS Read file function
DOS_BOOT	EQU	19H		; Boot vector
MS_DOS		EQU	21H		; MS-DOS vector
CMD_LINE	EQU	80H		; MS-DOS puts the command-line here
CR		EQU	0DH		; ASCII
LF		EQU	0AH		; ASCII
SCR		EQU	10H		; Screen interrupt
SET_PAGE	EQU	0500H		; Set active page 0
STK_SIZE	EQU	100H		; Size of load stack
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;	Define a couple of macros.
;
MKMSG	MACRO	LAB, MESG
	LOCAL	LAB, MESG
LAB	DB	"&MESG&"
	DB	CR,LF,0
ENDM
;
MKSTR	MACRO	LAB, MESG
	LOCAL	LAB, MESG
LAB	DB	"&MESG&"
	DB	0
ENDM
;
PUTS	MACRO	STRING
	MOV	SI,OFFSET STRING
	CALL	PUTSTR
ENDM
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;	This is the startup code. It reads the kernel and boot loader
;	into the proper memory locations and starts execution.
;
;	Note when the DOS loader calls this program, SS=DS=ES=CS. However
;	since SP is at the end of the segment, it must be put into a safe
;	place. Also, this processor will not acknowledge interrupts until
;	the instruction following any stack operations. It is therefore not
;	necessary (or proper) to alter the interrupt flag.
;
PSEG	SEGMENT PARA	PUBLIC 'TEXT'
	ASSUME	CS:PSEG, DS:PSEG, ES:PSEG, SS:PSEG
	ORG	0100H			; It's a COM file
MAIN	PROC	FAR
	MOV	AX,CS			; Get segment address
	ADD	AX,(OFFSET MAIN - PSEG) SHR 4	; Offset to 100H
	MOV	WORD PTR [LOD_SEG],AX	; Save load address
	ASSUME	DS:SCREEN, ES:SCREEN
	MOV	AX,SET_PAGE		; Set any screen card to page 0
	INT	SCR			; Screen control vector
	CLD
	MOV	BX,SCREEN		; Set to screen segment
	MOV	ES,BX			; Set segment
	MOV	WORD PTR ES:[PAGE_ONE],1234H	; See if it's writable
	PUSH	AX			; Exercise the bus
	POP	AX
	CMP	WORD PTR ES:[PAGE_ONE],1234H	; Was any RAM there?
	JNZ	NO_RAM			; Was no screen card
	MOV	DI,OFFSET PAGE_ONE	; Where to put the data
	XOR	SI,SI			; Move from DS:[0]
	MOV	CX,OFFSET END_TAG	; Data bytes to move
	INC	CX			; Maybe odd count
	SHR	CX,1			; Convert to words
	REP	MOVSW			; Relocate everything
	ADD	DI,STK_SIZE		; Offset for stack
	MOV	AX,OFFSET PAGE_ONE	; This will be CS:[0]
	SHR	AX,4			; Convert to paragraphs
	ADD	AX,BX			; Add screen segment
	MOV	ES,AX			; Set all segments
	MOV	DS,AX
	MOV	SS,AX			; Set this segment last
	MOV	SP,DI			; Set stack pointer
	PUSH	AX			; Push segment address
	PUSH	OFFSET RELOC		; Push starting location
	RET				; 'Return' to relocated code
	ASSUME	DS:PSEG, ES:PSEG
NO_RAM:	MOV	AX,CS
	MOV	ES,AX			; Restore segment
	MOV	AX,CS			; Get segment address
	ADD	AX,LAST_PARA		; Offset to end
	MOV	WORD PTR [LOD_SEG],AX	; Save load address
	JMP	START
MAIN	ENDP
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
	ASSUME	DS:PSEG, ES:PSEG

MKMSG <SignOn>		<JBOOT V1.0  Linux DOS boot loader>
MKSTR <BadFileMsg>	<Could not open Image file, >
MKMSG <CheckMsg>	<Not a boot-disk image>
MKMSG <ImageMsg>	<Bad image length>
MKMSG <DoneMsg>		< done>
MKMSG <BadLoc>		<Loaded too high in memory>
MKMSG <NoName>		<No boot file specified>
MKSTR <BootMsg>		<Boot sector: ... >
SetupMsg	DB	"loaded",13,10,"LINUX Setup: ... ",0
KernelMsg	DB	"loaded",13,10,"LINUX Kernel: ",0
LOD_SEG		DW	0		; Original load address
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;
START	PROC	NEAR
	MOV	SP,OFFSET MAIN		; Put SP out of the way
	PUTS	SignOn			; Print signon message
	MOV	AX,CS			; Get current code-segment
	ADD	AX,LAST_PARA		; Last paragraph used
	CMP	AX,KERNEL		; Check limits
	MOV	SI,OFFSET BadLoc	; Assume out-of range
	JA	BAD_EX			; Loaded too high
	JMP	SHORT DOIT		; We already printed signon message
RELOC:	MOV	SP,OFFSET MAIN		; Put SP out of the way
	PUTS	SignOn			; Print signon message
DOIT:	MOV	SI,CMD_LINE		; Command line
	LODSB				; Get bytes typed
	OR	AL,AL			; Anything entered?
	JNZ	WASCMD			; Yes
	MOV	SI,OFFSET NoName	; Point to error message
	JMP	SHORT BAD_EX		; Print string and exit
WASCMD:	CBW
	MOV	CX,AX			; Byte count
	JMP	SHORT BYPASS
CHECK:	LODSB
BYPASS:	CMP	BYTE PTR [SI], ' '	; Filter white space
	JA	CLEAN			; No more white space
	LOOP	CHECK			; Check whole command line
CLEAN:	MOV	BX,CX			; Need an index register
	MOV	BYTE PTR [SI+BX],0	; Make ASCIIZ
	MOV	DX,SI			; Point to file name
	MOV	AX,DOS_OPEN		; Open the specified file
	INT	MS_DOS			; Call MS-DOS
	JNC	OPENOK			; File open AX = handle
;
	PUTS	BadFileMsg		; Print the error message
	MOV	SI,DX			; Print the file name and exit
BAD_EX:	CALL	PUTSTR			; Write string
	MOV	AX,4C01H		; DOS exit function ERRORLEVEL=1
	INT	MS_DOS			; Call DOS

OPENOK:	MOV 	BX,AX			; Save file handle
	CALL	DISABLE			; Clear unused interrupt vectors
	PUTS	BootMsg			; Print the message
	XOR	DX,DX			; Offset 0
	MOV	DI,SETUP		; Target SETUP:0
	MOV	CX,OFFSET SETUP:KSTART	; Read one sector
	CALL	FREAD			; Read boot sector
	MOV	SI,OFFSET ImageMsg	; Point to error message
	JNZ	BAD_EX			; Bad length read
;
	CALL	BCHECK			; Check for valid boot sector
	MOV	SI,OFFSET CheckMsg	; Point to error message
	JNZ	BAD_EX			; Not a valid boot record
;
	PUTS	SetupMsg		; Print the message
	MOV	DX,OFFSET SETUP:KSTART	; Offset to write
	MOV	DI,SETUP		; Segment to write
	MOV	CX,SETUPLEN		; Bytes to read
	CALL	FREAD			; Read the data
	MOV	SI,OFFSET ImageMsg	; Point to error message
	JNZ	BAD_EX			; Not a valid image
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;	Copy the kernel image data to the area where we were first loaded.
;	We need to copy these data to memory we "own" when we use the file
;	system. We will move it into its final place after we close the file.
;
	PUTS	KernelMsg		; Print the message
	MOV	DI,WORD PTR [LOD_SEG]	; Where to put main image first
	MOV	CX,PAGE_LEN		; Read by pages
	XOR	DX,DX			; Offset zero
	XOR	BP,BP			; Use for the loop count
	SUB	DI,PAGE_SEG		; For following logic
READ:	ADD	DI,PAGE_SEG		; Next paragraph
	MOV	AX,0E2EH		; Write dots
	INT	SCR			; Screen BIOS
	INC	BP			; Keep track of reads
	CALL	FREAD			; Read the image
	JZ	READ			; Read whole image
;
	MOV	AX,DOS_CLOSE		; DOS Close function
	INT	MS_DOS			; Call DOS
	PUTS	DoneMsg			; Print the message
;
	MOV	CX,BP			; Loop counter
	MOV	DX,KERNEL		; Where to relocate the image
	MOV	BX,WORD PTR [LOD_SEG]	; Get where we put the data
	CMP	DX,BX			; Which way do we
	JNB	REVCPY			; Got to copy backwards
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;	We are loaded higher than "KERNEL" so we can copy normally.
;
COPYH:	MOV	AX,CX			; Save the copy count
	MOV	DS,BX			; Set source segment
	MOV	ES,DX			; Set destination segment
	MOV	CX,PAGE_LEN		; Copy by pages
	XOR	DI,DI			; Offset zero
	MOV	SI,DI			; Offset zero
	REP	MOVSB			; Copy the data
	ADD	BX,PAGE_SEG		; Set up new source
	ADD	DX,PAGE_SEG		; Set up new destination
	MOV	CX,AX			; Restore loop counter
	LOOP	COPYH
	JMP	SHORT MOVED		; All done okay
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;	We are loaded lower than "KERNEL" so we copy backwards to prevent
;	overwriting old data with new data before it's copied.
;
REVCPY:	STD				; Direction flag
	PUSH	DX			; Save source segment
	MOV	AX,PAGE_SEG		; Get page length in segments
	MOV	DX,CX			; Get number of pages
	DEC	DX			; If page = 1, no offset at all
	MUL	DX			; Times number of pages
	POP	DX			; Restore source segment
	ADD	BX,AX			; Calc starting destination page
	ADD	DX,AX			; Calc starting source page
COPYL:	MOV	AX,CX			; Save the copy count
	MOV	DS,BX			; Set source segment
	MOV	ES,DX			; Set destination segment
	MOV	CX,PAGE_LEN		; Copy by pages
	MOV	DI,PAGE_LEN - 1		; Offset page length (0 based)
	MOV	SI,PAGE_LEN - 1		; Offset page length (0 based)
	REP	MOVSB			; Copy the data
	SUB	BX,PAGE_SEG		; Set up new source
	SUB	DX,PAGE_SEG		; Set up new destination
	MOV	CX,AX			; Restore loop counter
	LOOP	COPYL
	CLD				; Restore direction flag
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;	Some people will have a memory manager installed when they attempt
;	to boot Linux. If they use the Microsoft memory manager, I can
;	fix the problems it will cause. The problem is that when the
;	linux setup portion attempts to go into protected mode, this
;	will cause an exception that is trapped by the memory manager.
;	The boot will fail.
;
;	Therefore, I replace the memory manager's saved boot vector
;	with the address of the relocated code. Then I "boot" the machine
;	using the interrupt 19h boot vector. This is trapped by the
;	memory manager which will enter protected mode, fix up everything
;	so the system can be booted, then attempt to boot the machine.
;	Since I have replaced its vector with the Linux entry point,
;	Linux will get control in "real" mode as required.
;
	ASSUME	DS:MEMMAN, ES:MEMMAN
MOVED:	MOV	AX,MEMMAN		; MS_DOS data segment
	MOV	DS,AX			; Set both segments
	MOV	ES,AX
	MOV	SI,OFFSET SAVED		; Memory manager interrupt table
	MOV	CX,8			; First 8 vectors
LOOK:	LODSB				; Get vector number
	OR	AL,AL			; End of the table
	JZ	ISTART			; Memory manager not installed
	CMP	AL,DOS_BOOT		; Find the boot interrupt number?
	JZ	FOUND			; Yes
	ADD	SI,4			; No, jump two words
	LOOP	LOOK			; Continue
	JMP	SHORT ISTART		; Memory manager not installed
FOUND:	MOV	DI,SI			; Point to offset location
	MOV	AX,OFFSET ISTART	; Get our starting location
	STOSW				; Replace theirs
	MOV	AX,CS			; Get our code segment
	STOSW				; Replace theirs
	INT	DOS_BOOT		; Now "boot" the machine!
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;
	ASSUME	DS:SETUP, ES:SETUP, SS:SETUP
ISTART:	MOV	AX,SETUP
	MOV	DS,AX
	MOV	ES,AX
	MOV	SS,AX
	MOV	SP,OFFSET KER_SP	; Set startup stack
	JMP	FAR PTR KSTART		; Do something more interesting
START	ENDP
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;	Tries to read CX bytes at DI:DX. Returns ZF=1 if okay. The
;	handle is already in BX.
;
FREAD	PROC	NEAR
	ASSUME	DS:NOTHING, ES:NOTHING
	PUSH	DS			; Save segment
	MOV	DS,DI			; Set segment
	MOV	AX,DOS_READ		; DOS read function
	INT	MS_DOS			; Call MS-DOS
	POP	DS			; Restore segment
	CMP	AX,CX			; Check for end-of-file
	RET
FREAD	ENDP
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;	This checks for a boot record signature
;
BCHECK	PROC	NEAR
	ASSUME	DS:SETUP, ES:NOTHING
	PUSH	DS
	MOV	AX,SETUP
	MOV	DS,AX
	CMP	WORD PTR [SIGNAT],0AA55H
	POP	DS
	RET
BCHECK	ENDP
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;	This prints ASCIIZ strings. Upon entry DS:SI points to the string.
;
PUTSTR	PROC	NEAR
	LODSB				; Get byte
	OR	AL,AL			; Check for the end
	JZ	PUTSEX			; It's all done
	MOV	AH,0EH			; Dumb terminal function
	INT	SCR			; Screen BIOS
	JMP	SHORT PUTSTR		; Continue
PUTSEX:	RET
PUTSTR	ENDP
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;	Two dummy interrupt service routines.
;
DUMMY0	PROC	FAR
	PUSH	AX			; Save register used
	MOV	AL,20H			; Non-specific end-of-interrupt
	OUT	20H,AL			; Send ACK to the controller
	POP	AX			; Restore register used
	IRET
DUMMY0	ENDP
;
DUMMY1	PROC	FAR
	PUSH	AX			; Save register used
	MOV	AL,20H			; Non-specific end-of-interrupt
	OUT	0A0H,AL			; Send ACK to the controller
	POP	AX			; Restore register used
	IRET
DUMMY1	ENDP
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;
;	We want to be able to boot Linux even if the user has been
;	executing MS-DOS programs and might even have a network installed.
;	Since we are going to load Linux at segment 1000h, we will certainly
;	overwrite any software existing there. This software might be using
;	interrupts. Therefore we disable any hardware interrupts that have
;	a vector above segment 1000h. We can't just turn off everything
;	because we have to use the disk drive to load the code. The load
;	may still fail if the network re-director is actually installed
;	because it may be in the overwritten location and it redirects MS-DOS
;	itself.
;
DISABLE	PROC	NEAR
	PUSH	BX			; Save file handle
	PUSH	DS			; Save segments
	PUSH	ES
	XOR	AX,AX			; Interrupt segment
	MOV	DS,AX			; Set segments to interrupt area
	MOV	ES,AX
	MOV	BX,KERNEL		; Get kernel segment
;
	MOV	SI,(8 SHL 2)		; First hardware interrupt table
	MOV	CX,8			; Vectors in this area
GETVTL:	MOV	DI,SI			; Possible starting vector
	LODSW				; Get offset
	SHR	AX,4			; Convert to paragraphs
	INC	AX			; Round up
	MOV	DX,AX			; Save
	LODSW				; Get segment
	ADD	AX,DX			; Sum paragraphs from the offset
	CMP	AX,BX			; Check location
	JB	VCTLOK			; Vector is okay
	CMP	AX,0F000H		; In the BIOS?
	JAE	VCTLOK			; It's okay
	MOV	AX,OFFSET DUMMY0	; Offset of dummy routine
	CLI
	STOSW
	MOV	AX,CS			; Segment of dummy routine
	STOSW
	STI
VCTLOK:	LOOP	GETVTL
;
;	Now check the second controller
;
	MOV	SI,(70H SHL 2)		; Second hardware interrupt table
	MOV	CX,8			; Vectors in this area
GETVTH:	MOV	DI,SI			; Possible vector
	LODSW				; Get offset
	SHR	AX,4			; Convert to paragraphs
	INC	AX			; Round up
	MOV	DX,AX			; Save
	LODSW				; Get segment
	ADD	AX,DX			; Sum paragraphs from the offset
	CMP	AX,BX			; Check location
	JB	VCTHOK			; Vector is okay
	CMP	AX,0F000H		; In the BIOS?
	JAE	VCTHOK			; It's okay
	MOV	AX,OFFSET DUMMY1	; Offset of dummy routine
	CLI
	STOSW
	MOV	AX,CS			; Segment of dummy routine
	STOSW
	STI
VCTHOK:	LOOP	GETVTH
	POP	ES			; Restore segments
	POP	DS
	POP	BX			; Restore file handle
	RET
DISABLE	ENDP
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
ALIGN	10H
LAST_PARA	EQU (($ - PSEG) SHR 4) + 1
END_TAG	LABEL	BYTE
PSEG		ENDS
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
	END MAIN

